package org.programmingbasics.iu.game;
import java.awt.*;
import java.awt.image.*;

/**
 * Used to allow rendering of graphics in an off-screen buffer to be
 * displayed later. (It essentially simulates a page flip, allowing you
 * to draw into one buffer while a second buffer is used to update the
 * screen/window.) This object is not fully synchronized, but it should
 * adequately support one thread drawing into the buffer with another
 * UI thread reading from a previously drawn version (with a safe "flip"
 * between the two versions). I'm not sure if it's strictly necessary,
 * but this RenderBuffer also enforces synchronous image copying since
 * I'm not sure whether Java correctly handles this.
 */
 
class RenderBuffer
{
	// The image buffers and their corresponding graphics handles
	//	
	BufferedImage [] buffers = new BufferedImage[2]; 
	Graphics2D [] g    = new Graphics2D[2];
	int width, height;

	// Which of the two buffers should be used for drawing into
	//
	int currentBuffer = 0;
	Graphics2D currentG;

	// Used to keep track of whether an asynchronous image copy has
	// completed or not
	//
	boolean imageDone;
	java.lang.Object imageLock = new java.lang.Object();
	ImageCallback imageCallback = new ImageCallback();

	// Used when images are asynchronously drawn--this gives us updates so 
	// that we know when an image has finished being drawn
	//
	class ImageCallback implements ImageObserver
	{
		public Image img;
		public int x;
		public int y;
		public Graphics2D g;
		public boolean imageUpdate(Image _img, int flags,
		       int _x, int _y, int width, int height)
		{
			// All the bits are available now, so redraw it
			//
			if ((flags & (FRAMEBITS|ALLBITS)) != 0) 
				g.drawImage( img, x, y, this );

			// If the image has finished drawing, then release any threads
			// waiting for the image drawing operation to finish
			//
			if ((flags & (ALLBITS|ABORT)) != 0)
			{
				synchronized (imageLock)
				{
					imageDone = true;
					imageLock.notifyAll();
				}
				return false;
			}
			
			return true;
		}
	}

	// Used when buffers are asynchronously copied--because Java uses
	// asynchronous image rendering, all sorts of synchronization 
	// messiness is needed to ensure that the copy occurred correctly
	//
	boolean copyDone;
	java.lang.Object copyLock = new java.lang.Object();
	CopyCallback copyCallback = new CopyCallback();
	class CopyCallback implements ImageObserver
	{
		public Image img;
		public int srcX1;
		public int srcY1;
		public int srcX2;
		public int srcY2;
		public int dstX1;
		public int dstY1;
		public int dstX2;
		public int dstY2;
		public Graphics g;
		
		public boolean imageUpdate(Image _img, int flags,
		       int x, int y, int width, int height)
		{
			// All the bits are available now, so redraw it (should we
			// stop accepting the ImageObserver thing since we're
			// redrawing the image?)
			//
			if ((flags & (FRAMEBITS|ALLBITS)) != 0)
				g.drawImage( img, dstX1, dstY1, dstX2, dstY2,
					srcX1, srcY1, srcX2, srcY2, this);

			// If the image has finished drawing, then release any threads
			// waiting for the image drawing operation to finish
			//
			if ((flags & (ALLBITS|ABORT)) != 0)
			{
				synchronized (copyLock)
				{
					copyDone = true;
					copyLock.notifyAll();
				}
				return false;
			}
			
			return true;
		}
	}	
   // Constructor for the render buffer. Requires a Component to create
   // images with.
   //
   public RenderBuffer( Component c, int w, int h )
   {
   	buffers[0] = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
   	buffers[1] = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
   
   	g[0] = buffers[0].createGraphics();
   	g[1] = buffers[1].createGraphics();
   
   	currentG = g[0];
   	width 	 = w;
   	height 	 = h;
   }
   // Synchronously copies the buffer to the specified graphics context
   //
   synchronized public void copyBuffer( Graphics g, 
   		int dstX1, int dstY1, int dstX2, int dstY2,
   		int srcX1, int srcY1, int srcX2, int srcY2)
   {
   	copyDone           = false;
   	copyCallback.srcX1 = srcX1;
   	copyCallback.srcY1 = srcY1;
   	copyCallback.srcX2 = srcX2;
   	copyCallback.srcY2 = srcY2;
   	copyCallback.dstX1 = dstX1;
   	copyCallback.dstY1 = dstY1;
   	copyCallback.dstX2 = dstX2;
   	copyCallback.dstY2 = dstY2;
   	copyCallback.img   = buffers[currentBuffer ^ 1];
   	copyCallback.g     = g;
   	
   	// Display the image
   	//
   	if (!g.drawImage( copyCallback.img, dstX1, dstY1, dstX2, dstY2,
   		srcX1, srcY1, srcX2, srcY2, copyCallback ))
   	{
   		// We weren't able to fully display the image, so wait until
   		// the callback tells us that the image has completed rendering
   		//
   		synchronized(copyLock)
   		{
   			while (!copyDone)
   			{
   				try {
   					copyLock.wait();
   				} catch (java.lang.InterruptedException e)
   				{
   				}				
   			}
   		}
   	}
   }
   // Synchronously draws an image into the image buffer
   //
   public void drawImage(Image img, int x, int y)
   {
   	imageDone         = false;
   	imageCallback.x   = x;
   	imageCallback.y   = y;
   	imageCallback.img = img;
   	imageCallback.g   = currentG;
   	
   	// Display the image
   	//
   	if (!currentG.drawImage( img, x, y, imageCallback ))
   	{
   		// We weren't able to fully display the image, so wait until
   		// the callback tells us that the image has completed rendering
   		//
   		synchronized(imageLock)
   		{
   			while (!imageDone)
   			{
   				try {
   					imageLock.wait();
   				} catch (java.lang.InterruptedException e)
   				{
   				}				
   			}
   		}
   	}
   }
   // Gets a graphics handle for drawing into the current buffer. The handle
   // should be released after it is used.
   //
   public Graphics2D getGraphics()
   {
   	return g[currentBuffer];
   }
   public int getHeight()
   {
   	return height;
   }
   public int getWidth()
   {
   	return width;
   }
   // Releases a graphics handle
   //
   public void releaseGraphics()
   {
   }
   // Used to switch which buffer is drawn into and which one should be
   // displayed (i.e. a page flip). Try to release all of your graphic
   // handles before calling this.
   //
   synchronized public void switchBuffer()
   {
   	currentBuffer ^= 1;
   	currentG = g[currentBuffer];
   }
}